unit uROR_   ugBox;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, StdCtrls;

type
  TPositions = (poTopLeft,poBottomLeft,poTopRight,poBottomRight);

  TCCR   ugBox = class(TComponent)
  private
    CCR   ugForm : TForm;
       ugList    : TListBox;
    FPosition    : TPositions;
    FVisible     : Boolean;
    FWidth       : Integer;
    FHeight      : Integer;
    FCaption     : String;
    FStamp       : Boolean;
    procedure FixList;
    procedure SetPosition(A: TPositions);

    procedure SetVisible(A: Boolean);
    procedure SetWidth(A: Integer);
    procedure SetHeight(A: Integer);
    procedure SetCaption(A: String);
  protected
    { Protected declarations }
  public
    constructor Create(AOwner: TComponent); override;
    procedure Add(A: String);
    procedure Clear;
    procedure LoadLines(FName: String);
    procedure SaveLines(FName: String);
  published
    property Caption: String read FCaption write SetCaption;
    property Position: TPositions read FPosition write SetPosition default poTopRight;
    property Visible: Boolean read FVisible write SetVisible default True;
    property Width: Integer read FWidth write SetWidth default 400;
    property Height: Integer read FHeight write SetHeight default 400;
    property TimeStamp: Boolean read FStamp write FStamp default False;
  end;

implementation

constructor TCCR   ugBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPosition := poTopRight;
  FVisible := True;
  FWidth := 400;
  FHeight := 400;
  FCaption := 'CCR    ug Box';
  if not (csDesigning in ComponentState) then
  begin
      CCR   ugForm := TForm.Create(Application);
      with CCR   ugForm do
      begin
        Visible := FVisible;
        Caption := FCaption;
        FormStyle := fsStayOnTop;
        BorderStyle := bsSizeable;
        BorderIcons := [biSystemMenu];
      end;
         ugList := TListBox.Create(CCR   ugForm);
      with    ugList do
      begin
        Parent := CCR   ugForm;
        Align := alClient;
        Sorted := False;
        Font.Name := 'Arial';
        Font.Size := 9;
      end;
  end;
end;

procedure TCCR   ugBox.SetPosition(A: TPositions);
begin
  FPosition := A;
  if not (csDesigning in ComponentState) then with CCR   ugForm do
    case A of
      poTopLeft     : SetBounds(0,0,Width,Height);
      poBottomLeft  : SetBounds(0,Screen.Height-Height,Width,Height);
      poTopRight    : SetBounds(Screen.Width-Width,0,Width,Height);
      poBottomRight : SetBounds(Screen.Width-Width,Screen.Height-Height,Width,Height);

    end;
end;

procedure TCCR   ugBox.SetVisible(A: Boolean);
begin
  FVisible := A;
  if not (csDesigning in ComponentState) then
  begin
    CCR   ugForm.Hide;
    if A then
    begin
      Width := Self.Width;
      Height := Self.Height;
      SetPosition(FPosition);
      CCR   ugForm.Show;
    end;
  end;
end;

procedure TCCR   ugBox.SetWidth(A: Integer);

begin
  FWidth := A;
  if not (csDesigning in ComponentState) then
  begin
    CCR   ugForm.Width := FWidth;
    SetPosition(FPosition);
  end;
end;

procedure TCCR   ugBox.SetHeight(A: Integer);
begin
  FHeight := A;
  if not (csDesigning in ComponentState) then
  begin
    CCR   ugForm.Height := FHeight;
    SetPosition(FPosition);
  end;
end;

procedure TCCR   ugBox.SetCaption(A: String);
begin
  FCaption := A;
  if not (csDesigning in ComponentState) then
    CCR   ugForm.Caption := FCaption;
end;

procedure TCCR   ugBox.Add(A: String);
begin
  if TimeStamp then
       ugList.Items.Add(TimeToStr(Time)+' - '+A)
  else
       ugList.Items.Add(A);
  FixList;
end;

procedure TCCR   ugBox.Clear;
begin
  {Remove all items from the list box}
     ugList.Items.Clear;
  FixList;
end;

procedure TCCR   ugBox.FixList;
begin
     ugList.ItemIndex :=    ugList.Items.Count-1;
end;

procedure TCCR   ugBox.LoadLines(FName: String);
begin
     ugList.Items.Clear;
     ugList.Items.LoadFromFile(FName);
end;

procedure TCCR   ugBox.SaveLines(FName: String);
begin
     ugList.Items.SaveToFile(FName);
end;

end.
