unit uROR_CommonUtils;
{
================================================================================
*
*       Package:        ROR - Clinical Case Registries
*       Date Created:   $Revision: 24 $  $Modtime: 4/12/07 3:55p $
*       Site:           Hines OIFO
*       Developers:
*                       Andrey Andriyevskiy
*                                      
*
*       Description:    Common Utils
*
*       Notes:
*
================================================================================
*       $Archive: /CCR v1.5/Current/uROR_CommonUtils.pas $
*
* $History: uROR_CommonUtils.pas $
 * 
 * *****************  Version 24  *****************
 * User: Vhaishgavris Date: 4/12/07    Time: 4:00p
 * Updated in $/CCR v1.5/Current
 * 
 * *****************  Version 23  *****************
 * User: Vhaishgavris Date: 3/08/07    Time: 1:47p
 * Updated in $/CCR v1.5/Current
 * 
 * *****************  Version 22  *****************
 * User: Vhaishgavris Date: 1/11/07    Time: 11:50a
 * Updated in $/CCR v1.5/Current
 * 
 * *****************  Version 21  *****************
 * User: Vhaishgavris Date: 6/29/06    Time: 8:47a
 * Updated in $/CCR v1.5/Current
 * 
 * *****************  Version 20  *****************
 * User: Vhaishgavris Date: 6/16/06    Time: 3:55p
 * Updated in $/CCR v1.5/Current
 * 
 * *****************  Version 19  *****************
 * User: Vhaishgavris Date: 9/07/05    Time: 3:45p
 * Updated in $/CCR v1.0/Current
 * 
 * *****************  Version 18  *****************
 * User: Vhaishgavris Date: 1/28/05    Time: 3:54p
 * Updated in $/CCR v1.0/Current
 * 
 * *****************  Version 17  *****************
 * User: Vhaishgavris Date: 10/15/04   Time: 10:17a
 * Updated in $/CCR v1.0/Current
 * 
 * *****************  Version 16  *****************
 * User: Vhaishgavris Date: 10/14/04   Time: 3:50p
 * Updated in $/CCR v1.0/Current
 * 
 * *****************  Version 15  *****************
 * User: Vhaishgavris Date: 8/06/04    Time: 4:08p
 * Updated in $/CCR v1.0/Current
 *
 * *****************  Version 14  *****************
 * User: Vhaishgavris Date: 6/14/04    Time: 3:54p
 * Updated in $/CCR v1.0/Current
 *
 * *****************  Version 13  *****************
 * User: Vhaishgavris Date: 5/07/04    Time: 3:55p
 * Updated in $/CCR v1.0/Current
*
*
================================================================================
}

interface
uses
  Classes,
  SysUtils,
  Dialogs,
  StdCtrls,
  Forms,
  Controls,
  comctrls,
  Graphics,
  extctrls;

procedure AccAnnotate(aWinControl: TWinControl);

function  FMDateToWindowsDate( FMDate: Double ): TDate;
function  FMDateTimeToWindowsDateTime( FMDateTime: Double ): TDateTime;

function  GetServerDateTime: TDateTime;

function  IntVal( str: String ): Integer;

function  WindowsDateTimeToFMDateTime( WinDate: TDateTime ): Double;
function  WindowsDateToFMDate( WinDate: TDate ): Double;

function FormatSSN(aSSN:String):String;
function ReFormatSSN(aSSN:String):String;

procedure UpdateListView(aStrings: TStrings;aLV: TListView;
  iCaptionIndex: Integer;anIndexes: array of integer; bClearList: boolean);

procedure AddParam(aName,aValue:String;sL:TStringList);

function StringByKey(aStrings:TStrings;aKey:String;aPosition:Integer):String;
procedure UpdateEdit(var Key: Char; Sender: TObject);

function RPCErrorCode(Results: TStrings = nil): Integer;
function CtrlColor(Active: Boolean): TColor;

implementation

uses
  uROR_Common, uROR_Utilities, MSAAConstants, WinApi.Windows;

procedure AccAnnotate(aWinControl: TWinControl);
var
  i: Integer;
  s: WideString;
  cc: TWinControl;
begin
  Exit; //??? Until the JAWS pecularity is sorted out

  if not Assigned(aWinControl) then
    Exit;

  with aWinControl do
    for i:=ControlCount-1 downto 0 do
      if Controls[i] is TWinControl then
        begin
          cc := TWinControl(Controls[i]);
          s := GetLongHint(cc.Hint);
          if s <> '' then
            if cc is TDateTimePicker then
              AccPropServices.SetHwndPropStr(cc.Handle, OBJID_CLIENT,
                CHILDID_SELF, PROPID_ACC_VALUE, PWideChar(s))
            else
              AccPropServices.SetHwndPropStr(cc.Handle, OBJID_CLIENT,
                CHILDID_SELF, PROPID_ACC_NAME, PWideChar(s));
          AccAnnotate(cc);
        end;
end;

function CtrlColor(Active: Boolean): TColor;
begin
  if Active then
    Result := clWindow
  else
    Result := clBtnFace;
end;

function IntVal(str: String): Integer;
begin
  if str <> '' then
    try
      Result := StrToInt(str);
    except
      on EConvertError do Result := 0;
      else raise;
    end
  else
    Result := 0;
end;

function GetServerDateTime: TDateTime;
var
  rawData: TStringList;
begin
  Result := Now;
  rawData := TStringList.Create;
  try
    if Broker.CallProc('ROR GET SERVER TIME', [], nil, rawData) then
      Result := FMDateTimeToWindowsDateTime(StrToFloat(rawData[1]))
  finally
    rawData.Free;
  end;
end;

function WindowsDateToFMDate(WinDate: TDate): Double;
var
  Year, Month, Day: Word;
begin
  DecodeDate(WinDate, Year, Month, Day);
  Result := ((Year - 1700) * 10000) + (Month * 100) + Day;
end;

function FMDateToWindowsDate(FMDate: Double): TDate;
var
  FMString: string;
  Year, Month, Day: Word;
begin
  FMString := FloatToStr(FMDate);
  try
    Year := StrToInt(Copy(FMString, 1, 3)) + 1700;
    Month := StrToInt(Copy(FMString, 4, 2));
    Day := StrToInt(Copy(FMString, 6, 2));
    Result := EncodeDate(Year, Month, Day);
  except
    on E: EConvertError do
      begin
        MessageDlg508('', 'Error Converting Date ' + FMString, mtError, [mbok], 0);
        Result := Now;
      end;
  else
    raise;
  end;
end;

function WindowsDateTimeToFMDateTime(WinDate: TDateTime): Double;
var
  Hour, Min, Sec, Msec: Word;
begin
  Result := WindowsDateToFMDate(WinDate);
  DecodeTime(WinDate, Hour, Min, Sec, Msec);
  Result := Result + (Hour * 0.01) + (Min * 0.0001) + (Sec * 0.000001);
end;

function FMDateTimeToWindowsDateTime(FMDateTime: Double): TDateTime;
var
  FMString: string;
  Hour, Min, Sec: Word;
begin
  FMString := FLoatToStr(FMDateTime);
  Result := FMDateToWindowsDate(FMDateTime);
  Hour := StrToIntDef(Copy(FMString, 9, 2), 0);
  Min := StrToIntDef(Copy(FMString, 11, 2), 0);
  Sec := StrToIntDef(Copy(FMString, 13, 2), 0);
  Result := Result + EncodeTime(Hour, Min, Sec, 0);
end;

function FormatSSN(aSSN:String):String;
begin
  try
    Result := Copy(aSSN,1,3)+'-'+Copy(aSSN,4,2)+'-'+Copy(aSSN,6,99);
  except
    Result := aSSN
  end;
end;

function ReFormatSSN(aSSN:String):String;
var
  i : Integer;
  sResult: String;
begin
  i := pos('-',aSSN);
  sResult := aSSN;
  while i <> 0 do
    begin
      sResult := copy(sResult,1,i-1)+copy(sResult,i+1,99);
      i := pos('-',sResult);
    end;
  Result := sResult;
end;

procedure UpdateListView(aStrings: TStrings;aLV: TListView;
  iCaptionIndex: Integer;anIndexes: array of integer; bClearList: boolean);
var
  s: String;
  li:TListItem;
  j,
  i,iCount: Integer;
begin
  try
    iCount := StrToInt(Piece(aStrings[0],'^',1));
    if iCount > aStrings.Count - 1 then
      iCount := aStrings.Count - 1;
  except
    iCount := 0;
  end;
  if bClearList then
    aLV.Items.Clear;
  for i := 1 to iCount do
  begin
    s := aStrings[i];
    li := aLV.Items.Add;
    li.Caption := piece(s,'^',iCaptionIndex);
    for j := Low(anIndexes) to High(anIndexes) do
      li.SubItems.Add(piece(s,'^',anIndexes[j]));
  end;
  try
    if aLV.Items.Count < 1 then exit;
    aLV.ItemFocused := aLV.Items[0];
    aLV.ItemFocused.Selected := True;
  except
  end;
end;

procedure AddParam(aName,aValue:String;sL:TStringList);
begin
  sL.Add(aName + '='+aValue);
end;

function RPCErrorCode(Results: TStrings): Integer;
var
  res: TStrings;
begin
  if Assigned(Results) then
    res := Results
  else
    res := Broker.Results;
  if res.Count > 0 then
    Result := StrToIntDef(Piece(Piece(res[0], '^'), '.'), -999)
  else
    Result := -999;
end;

function StringByKey(aStrings:TStrings;aKey:String;aPosition:Integer):String;
var
  s: String;
  i: Integer;
begin
  s := '';
  for i := 1 to aStrings.Count - 1 do
    if piece(aStrings[i],'^',aPosition) = aKey then
      s := aStrings[i];
  Result := s;
end;

procedure  UpdateEdit(var Key: Char; Sender: TObject);
var
  s,sY,sM: String;
begin
  if ((pos(Upcase(Key),'0123456789/') = 0))  then
    begin
      if Key <> Char(#13) then
        Key := Char(#8);
      exit;
    end;
  s := TEdit(Sender).Text;
  sY := piece(s,'/',2);
  sM := piece(s,'/',1);
  if ((s='') and (Key='/')) then
    begin
      Key := Char(#8);
      exit;
    end
  else if (pos('/',s)<>0) and (Key='/') then
    begin
      Key := Char(#0);
      exit;
    end
  else  if sY = '' then
    begin
      case Length(sM) of
        0:;
        1:
          begin
            if (sM<>'0') and (sM<>'1') then
              begin
                if Key <> '/' then
                  begin

                    TEdit(Sender).Text := '0'+sM +'/';
                  end
                else
                  TEdit(Sender).Text := '0'+sM;
                TEdit(Sender).SelStart := 3;
              end
            else if (sM = '0') and ((Key='/') or (Key='0')) then
              begin
                Key := Char(#0);
                exit;
              end;
          end;
        2:
          if (sM <> '11') and (sM <>'12') then
            begin
              if pos('0',sM) = 0 then
                begin
                  sM := '0'+copy(sM,1,1);
                  sY := copy(sM,2,4);
                end;
              TEdit(Sender).Text := sM +'/'+sY;
              TEdit(Sender).SelStart := 4;
              if Key='/' then Key:=Char(#0);
            end
          else
            begin
              if Key <>'/' then
                begin
                  if pos('/',s) <> 0 then
                    s := s +Key
                  else
                    s := s + '/'+Key
                end
              else
                s := s + '/';
              TEdit(Sender).Text := s;
              TEdit(Sender).SelStart := 4;
              Key := Char(#0);
            end;
        end;
    end
  else
    begin
      if (Key = Char(#8)) and
        (TEdit(Sender).SelStart = 3) then
        Key := Char(#0);
    end;
end;

end.








