unit uROR_ICRPatientData;
{
================================================================================
*
*       Package:        ROR - Clinical Case Registries
*       Date Created:   $Revision: 12 $  $Modtime: 3/06/07 2:16p $
*       Site:           Hines OIFO
*       Developers:     Andrey Andriyevskiy
*                                      
*       Description:    ICR Patient Object
*
*	Notes:
*
================================================================================
*	$Archive: /CCR v1.5/Current/uROR_ICRPatientData.pas $
*
* $History: uROR_ICRPatientData.pas $
 * 
 * *****************  Version 12  *****************
 * User: Vhaishgavris Date: 3/08/07    Time: 1:47p
 * Updated in $/CCR v1.5/Current
 * 
 * *****************  Version 11  *****************
 * User: Vhaishgavris Date: 8/19/05    Time: 3:51p
 * Updated in $/CCR v1.0/Current
 * 
 * *****************  Version 10  *****************
 * User: Vhaishgavris Date: 8/18/05    Time: 4:56p
 * Updated in $/CCR v1.0/Current
 * 
 * *****************  Version 8  *****************
 * User: Vhaishgavris Date: 2/16/05    Time: 3:55p
 * Updated in $/CCR v1.0/Current
 * 
 * *****************  Version 7  *****************
 * User: Vhaishgavris Date: 10/14/04   Time: 3:50p
 * Updated in $/CCR v1.0/Current
 *
*
================================================================================
}

interface

uses
  Classes, uROR_Common, uROR_PatientData, Dialogs;

type

  TICRPatientData = class(TPatientData)
  private
    fAIDSFirstDiag:             String;
    fAIDSOI:                    String;
    fAIDSOIDate:                String;
    fBloodComponent:            String;
    fBloodComponentFirst:       String;
    fBloodComponentLast:        String;
    fClottingFactor:            String;
    fClottingFactorOther:       String;
    fClottingFactorType:        String;
    fHCWorker:                  String;
    fHCWorkerOccupation:        String;
    fHSRwBisexualMale:          String;
    fHSRwHemofilia:             String;
    fHSRwHIV:                   String;
    fHSRwIVDrugUser:            String;
    fHSRwTransfusion:           String;
    fHSRwTransplant:            String;
    fNonprescrIVDrug:           String;
    fSexWithFemale:             String;
    fSexWithMale:               String;
    fTransplant:                String;

  protected
    procedure AssignRawData(Data: TStrings); override;
    procedure ClearPatientData; override;
    procedure PrepareRawData(Data: TStrings); override;
    procedure setAIDSFirstDiag(const Value: String);
    procedure setAIDSOI(const Value: String);
    procedure setAIDSOIDate(const Value: String);
    procedure setBloodComponent(const Value: String);
    procedure setBloodComponentFirst(const Value: String);
    procedure setBloodComponentLast(const Value: String);
    procedure setClottingFactor(const Value: String);
    procedure setClottingFactorOther(const Value: String);
    procedure setClottingFactorType(const Value: String);
    procedure setHCWorker(const Value: String);
    procedure setHCWorkerOccupation(const Value: String);
    procedure setHSRwBisexualMale(const Value: String);
    procedure setHSRwHemofilia(const Value: String);
    procedure setHSRwHIV(const Value: String);
    procedure setHSRwIVDrugUser(const Value: String);
    procedure setHSRwTransfusion(const Value: String);
    procedure setHSRwTransplant(const Value: String);
    procedure setNonprescrIVDrug(const Value: String);
    procedure setSexWithFemale(const Value: String);
    procedure setSexWithMale(const Value: String);
    procedure setTransplant(const Value: String);

  public
    property AIDSFirstDiag: String              read  fAIDSFirstDiag
                                                write setAIDSFirstDiag;
    property AIDSOI: String                     read  fAIDSOI
                                                write setAIDSOI;
    property AIDSOIDate: String                 read  fAIDSOIDate
                                                write setAIDSOIDate;
    property BloodComponent: String             read  fBloodComponent
                                                write setBloodComponent;
    property BloodComponentFirst: String        read  fBloodComponentFirst
                                                write setBloodComponentFirst;
    property BloodComponentLast: String         read  fBloodComponentLast
                                                write setBloodComponentLast;
    property ClottingFactor: String             read  fClottingFactor
                                                write setClottingFactor ;
    property ClottingFactorOther: String        read  fClottingFactorOther
                                                write setClottingFactorOther;
    property ClottingFactorType: String         read  fClottingFactorType
                                                write setClottingFactorType;
    property HCWorker: String                   read  fHCWorker
                                                write setHCWorker;
    property HCWorkerOccupation: String         read  fHCWorkerOccupation
                                                write setHCWorkerOccupation;
    property HSRwBisexualMale: String           read  fHSRwBisexualmale
                                                write setHSRwBisexualMale;
    property HSRwHemofilia: String              read  fHSRwHemofilia
                                                write setHSRwHemofilia;
    property HSRwHIV: String                    read  fHSRwHIV
                                                write setHSRwHIV;
    property HSRwIVDrugUser: String             read  fHSRwIVDrugUser
                                                write setHSRwIVDrugUser;
    property HSRwTransfusion: String            read  fHSRwTransfusion
                                                write setHSRwTransfusion;
    property HSRwTransplant: String             read  fHSRwTransplant
                                                write setHSRwTransplant;
    property NonprescrIVDrug: String            read  fNonprescrIVDrug
                                                write setNonprescrIVDrug;
    property SexWithFemale: String              read  fSexWithFemale
                                                write setSexWithFemale;
    property SexWithMale: String                read  fSexWithMale
                                                write setSexWithMale;
    property Transplant: String                 read  fTransplant
                                                write setTransplant;

    constructor Create(const aRegIEN: String);

  end;

implementation

uses
  SysUtils, uROR_Broker, uROR_CommonUtils, uROR_Utilities;

//////////////////////////////// TICRPatientData \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

constructor TICRPatientData.Create(const aRegIEN: String);
begin
  inherited;
  LoadRPCName   := 'RORICR PATIENT LOAD';
  SaveRPCName   := 'RORICR PATIENT SAVE';
end;

procedure TICRPatientData.AssignRawData(Data: TStrings);
var
  RawData: String;
begin
  inherited;

  //--- ICR Data
  RawData := Trim(StringByKey(Data, 'ICR', 1));
  fAIDSOI              := Piece(RawData, '^',  3);
  fAIDSOIDate          := Piece(RawData, '^',  4);
  //--- Piece 5 is reserved on the M side even though it isn't used
  fAIDSFirstDiag       := Piece(RawData, '^',  6);

  //--- Patient History
  RawData := Trim(StringByKey(Data, 'PH', 1));

  fSexWithMale         := Piece(RawData, '^',  3);
  fSexWithFeMale       := Piece(RawData, '^',  4);
  fNonprescrIVDrug     := Piece(RawData, '^',  5);
  fClottingFactor      := Piece(RawData, '^',  6);
  fClottingFactorType  := Piece(RawData, '^',  7);
  fClottingFactorOther := Piece(RawData, '^',  8);
  fHSRwIVDrugUser      := Piece(RawData, '^',  9);
  fHSRwBisexualMale    := Piece(RawData, '^', 10);
  fHSRwHemofilia       := Piece(RawData, '^', 11);
  fHSRwTransfusion     := Piece(RawData, '^', 12);
  fHSRwTransplant      := Piece(RawData, '^', 13);
  fHSRwHIV             := Piece(RawData, '^', 14);
  fBloodComponent      := Piece(RawData, '^', 15);
  fBloodComponentFirst := Piece(RawData, '^', 16);
  fBloodComponentLast  := Piece(RawData, '^', 17);
  fTransplant          := Piece(RawData, '^', 18);
  fHCWorker            := Piece(RawData, '^', 19);
  fHCWorkerOccupation  := Piece(RawData, '^', 20);
end;

procedure TICRPatientData.ClearPatientData;
begin
  inherited;
  fAIDSFirstDiag        := '';
  fAIDSOI               := '';
  fAIDSOIDate           := '';
  fBloodComponent       := '';
  fBloodComponentFirst  := '';
  fBloodComponentLast   := '';
  fClottingFactor       := '';
  fClottingFactorOther  := '';
  fClottingFactorType   := '';
  fHCWorker             := '';
  fHCWorkerOccupation   := '';
  fHSRwBisexualMale     := '';
  fHSRwHemofilia        := '';
  fHSRwHIV              := '';
  fHSRwIVDrugUser       := '';
  fHSRwTransfusion      := '';
  fHSRwTransplant       := '';
  fNonprescrIVDrug      := '';
  fSexWithFemale        := '';
  fSexWithMale          := '';
  fTransplant           := '';
end;

procedure TICRPatientData.PrepareRawData(Data: TStrings);
var
  RawData: String;
begin
  inherited;

  RawData := 'ICR^^' +
    fAIDSOI              + '^' +  //  3
    fAIDSOIDate          + '^' +  //  4
    '^' +                         //  5 (Reserved on M side)
    fAIDSFirstDiag;               //  6
  Data.Add(RawData);

  RawData := 'PH^^' +
    fSexWithMale         + '^' +  //  3
    fSexWithFeMale       + '^' +  //  4
    fNonprescrIVDrug     + '^' +  //  5
    fClottingFactor      + '^' +  //  6
    fClottingFactorType  + '^' +  //  7
    fClottingFactorOther + '^' +  //  8
    fHSRwIVDrugUser      + '^' +  //  9
    fHSRwBisexualMale    + '^' +  // 10
    fHSRwHemofilia       + '^' +  // 11
    fHSRwTransfusion     + '^' +  // 12
    fHSRwTransplant      + '^' +  // 13
    fHSRwHIV             + '^' +  // 14
    fBloodComponent      + '^' +  // 15
    fBloodComponentFirst + '^' +  // 16
    fBloodComponentLast  + '^' +  // 17
    fTransplant          + '^' +  // 18
    fHCWorker            + '^' +  // 19
    fHCWorkerOccupation;          // 20
  Data.Add(RawData);
end;

procedure TICRPatientData.setAIDSFirstDiag(const Value: String);
begin
  if Value <> fAIDSFirstDiag then
    begin
      fAIDSFirstDiag := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setAIDSOI(const Value: String);
begin
  if Value <> fAIDSOI then
    begin
      fAIDSOI := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setAIDSOIDate(const Value: String);
begin
  if Value <> fAIDSOIDate then
    begin
      fAIDSOIDate := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setBloodComponent(const Value: String);
begin
  if Value <> fBloodComponent then
    begin
      fBloodComponent := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setBloodComponentFirst(const Value: String);
begin
  if Value <> fBloodComponentFirst then
    begin
      fBloodComponentFirst := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setBloodComponentLast(const Value: String);
begin
  if Value <> fBloodComponentLast then
    begin
      fBloodComponentLast := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setClottingFactor(const Value: String);
begin
  if Value <> fClottingFactor then
    begin
      fClottingFactor := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setClottingFactorOther(const Value: String);
begin
  if Value <> fClottingFactorOther then
    begin
      fClottingFactorOther := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setClottingFactorType(const Value: String);
begin
  if Value <> fClottingFactorType then
    begin
      fClottingFactorType := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setHCWorker(const Value: String);
begin
  if Value <> fHCWorker then
    begin
      fHCWorker := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setHCWorkerOccupation(const Value: String);
begin
  if Value <> fHCWorkerOccupation then
    begin
      fHCWorkerOccupation := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setHSRwBisexualMale(const Value: String);
begin
  if Value <> fHSRwBisexualMale then
    begin
      fHSRwBisexualMale := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setHSRwHemofilia(const Value: String);
begin
  if Value <> fHSRwHemofilia then
    begin
      fHSRwHemofilia := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setHSRwHIV(const Value: String);
begin
  if Value <> fHSRwHIV then
    begin
      fHSRwHIV := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setHSRwIVDrugUser(const Value: String);
begin
  if Value <> fHSRwIVDrugUser then
    begin
      fHSRwIVDrugUser := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setHSRwTransfusion(const Value: String);
begin
  if Value <> fHSRwTransfusion then
    begin
      fHSRwTransfusion := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setHSRwTransplant(const Value: String);
begin
  if Value <> fHSRwTransplant then
    begin
      fHSRwTransplant := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setNonprescrIVDrug(const Value: String);
begin
  if Value <> fNonprescrIVDrug then
    begin
      fNonprescrIVDrug := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setSexWithFemale(const Value: String);
begin
  if Value <> fSexWithFeMale then
    begin
      fSexWithFeMale := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setSexWithMale(const Value: String);
begin
  if Value <> fSexWithMale then
    begin
      fSexWithMale := Value;
      Modified := True;
    end;
end;

procedure TICRPatientData.setTransplant(const Value: String);
begin
  if Value <> fTransplant then
    begin
      fTransplant := Value;
      Modified := True;
    end;
end;

end.
